// This file was generated by gir (https://github.com/gtk-rs/gir)
// from
// from gir-files (https://github.com/gtk-rs/gir-files)
// DO NOT EDIT

use crate::ffi;
use glib::GStr;

#[cfg(feature = "v26_1")]
#[cfg_attr(docsrs, doc(cfg(feature = "v26_1")))]
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_CHECK")]
pub static ANNOT_TEXT_ICON_CHECK: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_CHECK) };
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_CIRCLE")]
pub static ANNOT_TEXT_ICON_CIRCLE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_CIRCLE) };
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_COMMENT")]
pub static ANNOT_TEXT_ICON_COMMENT: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_COMMENT) };
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_CROSS")]
pub static ANNOT_TEXT_ICON_CROSS: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_CROSS) };
#[cfg(feature = "v26_1")]
#[cfg_attr(docsrs, doc(cfg(feature = "v26_1")))]
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_CROSS_HAIRS")]
pub static ANNOT_TEXT_ICON_CROSS_HAIRS: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_CROSS_HAIRS) };
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_HELP")]
pub static ANNOT_TEXT_ICON_HELP: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_HELP) };
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_INSERT")]
pub static ANNOT_TEXT_ICON_INSERT: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_INSERT) };
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_KEY")]
pub static ANNOT_TEXT_ICON_KEY: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_KEY) };
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_NEW_PARAGRAPH")]
pub static ANNOT_TEXT_ICON_NEW_PARAGRAPH: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_NEW_PARAGRAPH) };
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_NOTE")]
pub static ANNOT_TEXT_ICON_NOTE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_NOTE) };
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_PARAGRAPH")]
pub static ANNOT_TEXT_ICON_PARAGRAPH: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_PARAGRAPH) };
#[cfg(feature = "v26_1")]
#[cfg_attr(docsrs, doc(cfg(feature = "v26_1")))]
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_RIGHT_ARROW")]
pub static ANNOT_TEXT_ICON_RIGHT_ARROW: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_RIGHT_ARROW) };
#[cfg(feature = "v26_1")]
#[cfg_attr(docsrs, doc(cfg(feature = "v26_1")))]
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_RIGHT_POINTER")]
pub static ANNOT_TEXT_ICON_RIGHT_POINTER: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_RIGHT_POINTER) };
#[cfg(feature = "v26_1")]
#[cfg_attr(docsrs, doc(cfg(feature = "v26_1")))]
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_STAR")]
pub static ANNOT_TEXT_ICON_STAR: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_STAR) };
#[cfg(feature = "v26_1")]
#[cfg_attr(docsrs, doc(cfg(feature = "v26_1")))]
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_UP_ARROW")]
pub static ANNOT_TEXT_ICON_UP_ARROW: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_UP_ARROW) };
#[cfg(feature = "v26_1")]
#[cfg_attr(docsrs, doc(cfg(feature = "v26_1")))]
#[doc(alias = "POPPLER_ANNOT_TEXT_ICON_UP_LEFT_ARROW")]
pub static ANNOT_TEXT_ICON_UP_LEFT_ARROW: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::POPPLER_ANNOT_TEXT_ICON_UP_LEFT_ARROW) };
